#pragma once
#include <map>
#include <unordered_set>
#include <vector>
#include <set>
#include "Exception.h"
#include "Types.h"
#include "Print_Functions.h"
#include <algorithm>
#include <cstdlib>


template<uint8_t DIM>
class MeshLoader
{
protected:
	std::vector<Node<DIM>> nodes;
	std::vector<FiniteElement> fin_els;
	std::multimap<uint32_t, BoundaryFiniteElement> b_fin_els;
	

	//Load mesh from file
	virtual void loadMesh(const std::string& filename) = 0;


	//Adds node to "nodes" container
	void AddNode(const Node<DIM>& cur_node)
	{
		nodes.push_back(cur_node);
	}


	//Adds FE to "fe_els" container
	void AddFE(const FiniteElement& cur_fe)
	{
		fin_els.push_back(cur_fe);
	}

	
	//Adds BFE to "fe_els" container
	void AddBFE(const uint32_t& bound_id, const BoundaryFiniteElement& cur_bfe)
	{
		b_fin_els.emplace(bound_id, cur_bfe);
	}
	

private:

	Node<DIM> createMiddleNode(const Node<DIM>& node1, const Node<DIM>& node2)
	{
		Node<DIM> new_node;

		new_node.id = nodes.size() + 1;

		for (uint8_t i = 0; i < DIM; ++i)
		{
			new_node.coord[i] = (node1.coord[i] + node2.coord[i]) / 2;
		}

        // in_node    bool         
        //  if else    
        // new_node.in_node = node1.in_node && node2.in_node;
		new_node.in_node = node1.in_node && node2.in_node;

		return new_node;
	}


public:

	MeshLoader() {};

	//Returns nodes container
	const std::vector<Node<DIM>>& getNodeContainer() const noexcept { return nodes; }


	//Returns finite elements container
	const std::vector<FiniteElement>& getFEContainer() const noexcept { return fin_els; }


	//Returns boundary finite elements container
	const std::multimap<uint32_t, BoundaryFiniteElement>& getBFEContainer() const noexcept { return b_fin_els; }

	
	//Searches finite elements with nodes id1, id2, id3
	std::vector<FiniteElement> findFE(const uint32_t& id1, const uint32_t& id2, const uint32_t& id3) const
	{
		std::vector<FiniteElement> buffer;
		std::vector<FiniteElement>::const_iterator cur_fe_it = fin_els.begin();
		uint8_t fe_vertex_cnt = fin_els.begin()->nodes.size(); //we take size of finite element vector "nodes"(we take it from first element just for simplicity)


		while (cur_fe_it != fin_els.end())
		{
			cur_fe_it = std::find_if(cur_fe_it, fin_els.end(),
			[id1, id2, id3, fe_vertex_cnt](const FiniteElement& ordered_fe)
			{
				uint8_t cnt = 0; //amount of matched nodes
				for (uint8_t i = 0; i < fe_vertex_cnt; ++i)
				{
					if ((ordered_fe.nodes[i] == id1) || (ordered_fe.nodes[i] == id2) || (ordered_fe.nodes[i] == id3))
					{
						cnt++;
					}
				}
				return cnt == 3; //if "ordered_fe" has nodes id1, id2, id3 then returned value is "true"
			});

			if (cur_fe_it != fin_els.end())
			{
				buffer.push_back(*cur_fe_it);
				cur_fe_it++;
			}
		}

		if (buffer.size() == 0)
			throw EmptyBufferExc();
		else
		{
			//PRINT FOUNDED ELEMENTS
			printFEs(buffer.begin(), buffer.end());
		}

		return buffer;
	}


	//Searches finite elements with nodes id1, id2
	std::vector<FiniteElement> findFE(const uint32_t& id1, const uint32_t& id2) const
	{
		std::vector<FiniteElement> buffer;
		std::vector<FiniteElement>::const_iterator cur_fe_it = fin_els.begin();
		uint8_t fe_vertex_cnt = fin_els.begin()->nodes.size(); //we take size of finite element vector "nodes"(we take it from first element just for simplicity)


		while (cur_fe_it != fin_els.end())
		{
			cur_fe_it = std::find_if(cur_fe_it, fin_els.end(),
				[id1, id2, fe_vertex_cnt](const FiniteElement& ordered_fe)
			{
				uint8_t cnt = 0; //amount of matched nodes
				for (uint8_t i = 0; i < fe_vertex_cnt; ++i)
				{
					if ((ordered_fe.nodes[i] == id1) || (ordered_fe.nodes[i] == id2))
					{
						cnt++;
					}
				}
				return cnt == 2; //if "ordered_fe" has nodes id1, id2 then returned value is "true"
			});

			if (cur_fe_it != fin_els.end())
			{
				buffer.push_back(*cur_fe_it);
				cur_fe_it++;
			}
		}

		if (buffer.size() == 0)
			throw EmptyBufferExc();
		else
		{
			//PRINT FOUNDED ELEMENTS
			printFEs(buffer.begin(), buffer.end());
		}

		return buffer;
	}


	//Returns container with nodes on boundary "boundary_id"(   ,    "boundary_id")
	std::vector<Node<DIM>> getBoundaryNodes(const uint32_t boundary_id)
	{
		std::set<Node<DIM>> buffer_set; //this container is additional(to avoid nodes dublicates)

		auto range = b_fin_els.equal_range(boundary_id); //"range" is pair of iterators pointing on begin and end of range of elements with key "boundary_id"

		for (auto it = range.first; it != range.second; ++it)
		{
			for (uint8_t i = 0; i < it->second.nodes.size(); ++i)
			{
				buffer_set.insert(getNodeWithID(it->second.nodes[i]));
			}
		}
		
		if (buffer_set.size() == 0)
			throw EmptyBufferExc();
		else
		{
			/*
			for (const auto& buf_el : buffer_set)
			{
                //    ,   
                //   ,   buffer_vec
                //      
                //  ,      
				*std::back_inserter(buffer_vec) = buf_el;
			}
			*/
			std::vector<Node<DIM>> buffer_vec(buffer_set.begin(), buffer_set.end());

			//PRINT FOUNDED ELEMENTS
			printNodes<DIM>(buffer_vec.begin(), buffer_vec.end());
		
			return buffer_vec;
		}
	}

	
	//Returns FE container with id "mater_id" of material
	std::vector<FiniteElement> getMaterialFEs(const uint16_t&  mater_id) const 
	{
		std::vector<FiniteElement> buffer;


		for (auto it = fin_els.begin(); it != fin_els.end(); ++it)
		{
			if (it->material_id == mater_id)
			{
				buffer.push_back(*it);
			}
		}

		if (buffer.size() == 0)
			throw EmptyBufferExc();
		else
		{
			//PRINT FOUNDED ELEMENTS
			printFEs(buffer.begin(), buffer.end());
		}

		return buffer;
	}
	

	//Returns container with BFEs with boundary id "boundary_id"(    ,    "boundary_id")
	std::vector<BoundaryFiniteElement> getBoundaryBFEs(const uint32_t boundary_id) const
	{
		auto range = b_fin_els.equal_range(boundary_id);
		std::vector<BoundaryFiniteElement> buffer;
		
		
        //     
		for (auto it = range.first; it != range.second; ++it)
			buffer.push_back(it->second);
		

		if (buffer.empty())
		{
			throw EmptyBufferExc();
		}
		else
		{
			//PRINT ALL ELEMENTS
			printBFEs(buffer.begin(), buffer.end());
		}

		return buffer;
	}


	//Prints BFEs with boundary id "boundary_id"(  ,    "boundary_id")
	void printBoundaryBFEs(const uint32_t boundary_id) const
	{
		uint8_t bfe_vertex_cnt = b_fin_els.begin()->second.nodes.size(); //we take size of boundary finite element vector "nodes"(we take it from 1-st container element just for simplicity)
		auto range = b_fin_els.equal_range(boundary_id);
		std::vector<BoundaryFiniteElement> buffer;


		for (auto it = range.first; it != range.second; ++it)
			buffer.push_back(it->second);


		if (buffer.empty())
		{
			throw EmptyBufferExc();
		}
		else
		{
			//PRINT ALL ELEMENTS
			std::cout << "\n+------------+-------------+--";
			for (uint8_t i = 0; i < bfe_vertex_cnt; ++i) std::cout << "------------";
			std::cout << "+";
			std::cout << "\n|     ID     | Boundary ID |  NODES       ";
			for (uint8_t i = 0; i < bfe_vertex_cnt - 1; ++i) std::cout << "            ";
			std::cout << "|";
			std::cout << "\n+------------+-------------+--";
			for (uint8_t i = 0; i < bfe_vertex_cnt; ++i) std::cout << "------------";
			std::cout << "+";
			
			std::ranges::copy(buffer, std::ostream_iterator<const BoundaryFiniteElement&>(std::cout));
		}
	}


	//Returns container of vectors with neighbours of every node(i.e. resultcontainer[i] is vector of neighbours of node with id "i").
	//Indexation in result container starts from 1!!!
	std::vector<std::vector<Node<DIM>>> getNodesNeighbours()
	{
		std::vector<std::vector<Node<DIM>>> buffer_vec; //buffer_vec[i] is vector of neighbours of node with ID "i" 
		std::vector<std::set<Node<DIM>>> buffer_set; //buffer_set[i] = buffer_vec[i](this container is to avoid dublications) 
		uint8_t fe_vertex_cnt = fin_els.begin()->nodes.size(); //we take size of finite element vector "nodes"(we take it from first element just for simplicity)
		uint8_t bfe_vertex_cnt = b_fin_els.begin()->second.nodes.size(); //we take size of boundary finite element vector "nodes"(we take it from first element just for simplicity)
		
		buffer_set.resize(nodes.size() + 1); //+1 because indexation starts from 1
		buffer_vec.resize(nodes.size() + 1); //+1 because indexation starts from 1
		

		for (auto it = fin_els.begin(); it != fin_els.end(); ++it)
		{
			for (uint8_t i = 0; i < fe_vertex_cnt; ++i)
			{
				for (uint8_t j = 0; j < fe_vertex_cnt; ++j)
				{
					if (i != j)
					{
						buffer_set[it->nodes[i]].insert(getNodeWithID(it->nodes[j]));
					}
				}
			}
		}

		for (auto it = b_fin_els.begin(); it != b_fin_els.end(); ++it)
		{
			for (uint8_t i = 0; i < bfe_vertex_cnt; ++i)
			{
				for (uint8_t j = 0; j < bfe_vertex_cnt; ++j)
				{
					if (i != j)
					{
						buffer_set[it->second.nodes[i]].insert(getNodeWithID(it->second.nodes[j]));
					}
				}
			}
		}

		
		for (uint32_t i = 1; i < buffer_set.size(); ++i)
		{
			/*
            //     vector::insert
            //  ,   
			for (auto& cur_vec_el : buffer_set[i])
			{
				*std::back_inserter(buffer_vec[i]) = cur_vec_el;
			}
			*/
			buffer_vec[i].assign(buffer_set[i].begin(), buffer_set[i].end());
		}
		

		
	
		return buffer_vec;
	}


	//Returns Node with given ID from "nodes" container
	Node<DIM>& getNodeWithID(const uint32_t node_id)
	{

		auto node = std::find_if(nodes.begin(), nodes.end(),
                                                    // node_id    
													[node_id](const Node<DIM>& cur_node)
													{return cur_node.id == node_id;});

		if (node == nodes.end()) throw NodeWrongIDExc();
        //   
        //      :
        //        
		else return *node;
	}


	void insertInCentres()
	{
		std::unordered_set<Edge, HashEdge> edges; //edges container

		for (auto& cur_fe : fin_els) //go through container of finite elements
		{
			std::vector<uint32_t> cur_fe_nodes = cur_fe.nodes; //current finite element nodes vector

			//Go through all edges of current finite element
			for (auto i = cur_fe_nodes.begin(); i != cur_fe_nodes.end() - 1; ++i)
			{
				for (auto j = i + 1; j != cur_fe_nodes.end(); ++j)
				{
					Edge cur_edge(*i, *j);

					auto cur_edge_it = edges.find(cur_edge); //find edge "cur_edge" in container of edges(this container contains edges that we have found before)

					if (cur_edge_it == edges.end()) //edge was not processed before
					{
						//Adding in nodes container
						Node<DIM> new_node = createMiddleNode(getNodeWithID(*i), getNodeWithID(*j));
						nodes.push_back(new_node);

						//Adding in edges container
						cur_edge.central_node_id = new_node.id;
						edges.insert(cur_edge);

						//Adding in current FE nodes vector
						cur_fe.nodes.push_back(cur_edge.central_node_id);
					}
					else
					{
						//Adding in current FE nodes vector
						cur_fe.nodes.push_back(cur_edge_it->central_node_id);
					}
				}
			}
		}

		for (auto& cur_bfe : b_fin_els) //go through container of boundary finite elements
		{
			std::vector<uint32_t> cur_bfe_nodes = cur_bfe.second.nodes;  //current boundary finite element nodes vector


			//Go through all edges of current finite element
			for (auto i = cur_bfe_nodes.begin(); i != cur_bfe_nodes.end() - 1; ++i)
			{
				for (auto j = i + 1; j != cur_bfe_nodes.end(); ++j)
				{
					Edge cur_edge(*i, *j);
					auto cur_edge_it = edges.find(cur_edge); //find edge "cur_edge" in container of edges(this container contains edges that we have found before)

					
					if (cur_edge_it == edges.end())
					{
						//Adding in nodes container
						Node<DIM> new_node = createMiddleNode(getNodeWithID(*i), getNodeWithID(*j));
						nodes.push_back(new_node);

						//Adding in edges container
						cur_edge.central_node_id = new_node.id;
						edges.insert(cur_edge);

						//Adding in current BFE nodes vector
						cur_bfe.second.nodes.push_back(cur_edge.central_node_id);
					}
					else
					{
						//Adding in current BFE nodes vector
						cur_bfe.second.nodes.push_back(cur_edge_it->central_node_id);
					}
				}
			}
		}
	}


	//Returns size of container "nodes"
	uint32_t getNodeCnt()
	{
		return nodes.size();
	}


	//Returns size of container "fin_els"
	uint32_t getFECnt()
	{
		return fin_els.size();
	}


	//Returns size of container "b_fin_els"
	uint32_t getBFECnt()
	{
		return b_fin_els.size();
	}
};
