#pragma once


#include <iostream>
#include <vector>
#include <array>
#include <fstream>
#include <functional>
#include <iomanip>
//#include <stdint.h> //VS2012


//--------------------------------------Node--------------------------------------

template <uint8_t DIM> //space dimension
struct Node
{
	uint32_t id; //node id
	std::array<double, DIM> coord; //node coordinates in DCS
	bool in_node; //true: inner node, false: node is FE(finite element) vertex

	Node(uint32_t id_val = 0, bool in_node_val = false); //constructor

	bool operator < (const Node<DIM>& comp_node) const;

	//With "friend" it does not work
	//friend std::ifstream& operator>>(std::ifstream& file, Node<DIM>& cur_node); //input from file
};


//*******************************Functions*********************************************

//Constructor
template<uint8_t DIM>
Node<DIM>::Node(uint32_t id_val, bool in_node_val) : id(id_val), in_node(in_node_val)
{
	for (uint8_t i = 0; i < DIM; ++i)
	{
		coord[i] = 0.0L;
	}
}


//Input from file
template<uint8_t DIM>
std::ifstream& operator>>(std::ifstream& file, Node<DIM>& cur_node)
{
	for (uint8_t j = 0; j < DIM; ++j)
	{
		file >> cur_node.coord[j];
	}
	return file;
}


template <uint8_t DIM>
bool Node<DIM>::operator < (const Node<DIM>&comp_node) const { return id < comp_node.id; }


//--------------------------------------FiniteElement--------------------------------------

struct FiniteElement
{
	uint32_t id; //FE id
	uint16_t material_id; //FE material id
	std::vector<uint32_t> nodes; //FE vertex ides

	
	FiniteElement(); //constructor

	//With "friend" it does not work
	//friend std::ifstream& operator>>(std::ifstream& file, FiniteElement<FEVertexCnt>& cur_fe); //input from file
};


//*******************************Functions*********************************************

//Constructor
FiniteElement::FiniteElement()
{
	constexpr uint16_t fe_vert_cnt = 4; //amount of vertexes in every finite element
	id = 0;
	material_id = 0;
	nodes.resize(fe_vert_cnt);

	for (uint8_t i = 0; i < nodes.size(); ++i)
	{
		nodes[i] = 0;
	}
}


//Input from file
std::ifstream& operator>>(std::ifstream& file, FiniteElement& cur_fe)
{
	file >> cur_fe.material_id;
	for (uint8_t i = 0; i < cur_fe.nodes.size(); ++i)
	{
		file >> cur_fe.nodes[i];
	}

	return file;
}



//--------------------------------------BoundaryFiniteElement--------------------------------------

struct BoundaryFiniteElement
{
	uint32_t id; //FE id
	uint32_t boundary_id; //BFE boundary id
	std::vector<uint32_t> nodes; //BFE boundary vertex ides
	

	BoundaryFiniteElement(); //constructor
	
	//With "friend" it does not work
	//friend std::ifstream& operator>>(std::ifstream& file, BoundaryFiniteElement<BFEBoundaryVertexCnt>& cur_bfe); //input from file
};


//*******************************Functions*********************************************

//Constructor
BoundaryFiniteElement::BoundaryFiniteElement()
{
	constexpr uint16_t bfe_vert_cnt = 3; //amount of vertexes in every boundary finite element
	id = 0;
	boundary_id = 0;
	nodes.resize(bfe_vert_cnt);
	for (uint8_t i = 0; i < nodes.size(); ++i)
	{
		nodes[i] = 0;
	}
}


//Input from file
std::ifstream& operator>>(std::ifstream& file, BoundaryFiniteElement& cur_bfe)
{
	file >> cur_bfe.boundary_id;
	for (uint8_t i = 0; i < cur_bfe.nodes.size(); ++i)
	{
		file >> cur_bfe.nodes[i];
	}

	return file;
}


//Output in stream
std::ostream& operator<<(std::ostream& p_out, const BoundaryFiniteElement& cur_bfe)
{
	p_out << "\n| " << std::setw(10) << cur_bfe.id;
	p_out << " | " << std::setw(11) << cur_bfe.boundary_id << " |  ";
	for (uint8_t i = 0; i < cur_bfe.nodes.size() - 1; ++i) p_out << std::setw(10) << cur_bfe.nodes[i] << ", ";
	p_out << std::setw(10) << cur_bfe.nodes[cur_bfe.nodes.size() - 1] << "  | ";
	p_out << "\n+------------+-------------+--";
	for (uint8_t i = 0; i < cur_bfe.nodes.size(); ++i) p_out << "------------";
	p_out << "+";

	return p_out;
}


//--------------------------------------Edge--------------------------------------


struct Edge
{
	uint32_t node1_id;
	uint32_t node2_id;
	uint32_t central_node_id;


	Edge() : node1_id(0), node2_id(0), central_node_id(0) {}


	Edge(uint32_t n1_id, uint32_t n2_id) : node1_id(n1_id), node2_id(n2_id)
	{
		central_node_id = 0;
	}

	bool operator ==(const Edge& edge2) const noexcept
	{
		return (((node1_id == edge2.node1_id) && (node2_id == edge2.node2_id)) || ((node1_id == edge2.node2_id) && (node2_id == edge2.node1_id)));
	}
};



//--------------------------------------HashEdge--------------------------------------
struct HashEdge
{
	size_t operator()(const Edge& edge) const
	{
		size_t seed = 0;
		seed ^= std::hash<uint32_t>()(edge.node1_id) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
		seed ^= std::hash<uint32_t>()(edge.node2_id) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
	
		return seed;
	}
};
